import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

import django
from django.db import models
from django.db.models import Count, Q
from django.db.models.functions import Concat

# other imports
from collections.abc import Sequence
import re
import os
import datetime
import time
import numpy as np
import itertools
from astropy.coordinates import SkyCoord
from pathlib import Path
import astropy.io.fits as fits
import glob
from astropy.time import Time

# iop4 imports

from iop4lib.enums import *
from iop4lib.telescopes import Telescope
from iop4lib.instruments import Instrument
from .fields import FlagChoices, FlagBitField
from iop4lib.utils import  get_mem_parent_from_child, get_total_mem_from_child, get_mem_current, get_mem_children
from iop4lib.utils.parallel import epoch_bulkreduce_multiprocesing

# logging

import logging
logger = logging.getLogger(__name__)

from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from iop4lib.db import RawFit, ReducedFit, Epoch
    
class Epoch(models.Model):
    """A class representing an epoch.
    
    Identified by the telescope and date of the night. Provides method for fetching the data
    from the telescope archives and reducing the data.
    """

    # Database fields and information

    # identifiers 
    
    telescope = models.CharField(max_length=255, choices=TELESCOPES.choices)
    night = models.DateField()

    # flags

    class FLAGS(FlagChoices):
        ERROR = 1 << 1
        ERROR_LISTING = 1 << 2
        ERROR_LISTING_REMOTE = 1 << 3
        ERROR_LISTING_LOCAL = 1 << 4
        LISTED = 1 << 5
        
    flags = FlagBitField(choices=FLAGS.choices(), default=0)

    # db info and constraints

    class Meta:
        app_label = 'iop4api'
        verbose_name = "Epoch"
        verbose_name_plural = "Epochs"
        constraints = [
            models.UniqueConstraint(fields=['telescope', 'night'], name='telescope_night_unique')
        ]
    
    # properties of epoch

    @property
    def epochname(self):
        """Returns a human readable string that uniquely identifies the epoch, as an alternative to the id."""
        return f"{self.telescope}/{self.night.strftime('%Y-%m-%d')}"

    @property
    def rawfitsdir(self):
        """Returns the path to the directory where the raw fits of this epoch are stored."""
        return os.path.join(iop4conf.datadir, "raw", self.epochname)
    
    @property
    def calibrationdir(self):
        """Returns the path to the directory where the calibration files of this epoch are stored."""
        return os.path.join(iop4conf.datadir, "calibration", self.epochname)
    
    @property
    def masterbiasdir(self):
        """Returns the path to the directory where the masterbias files of this epoch are stored."""
        return os.path.join(iop4conf.datadir, "masterbias", self.epochname)

    @property
    def masterflatdir(self):
        """Returns the path to the directory where the masterflat files of this epoch are stored."""
        return os.path.join(iop4conf.datadir, "masterflat", self.epochname)

    @property 
    def masterdarkdir(self):
        """Returns the path to the directory where the masterdark files of this epoch are stored."""
        return os.path.join(iop4conf.datadir, "masterdark", self.epochname)

    @property
    def yyyymmdd(self):
        """Returns the date of the epoch in the format YYYYMMDD."""
        return self.night.strftime("%Y%m%d")
    
    @property
    def yymmdd(self):
        """Returns the date of the epoch in the format YYMMDD."""
        return self.night.strftime("%y%m%d")
    
    # helper property to return the jyear of the epoch at noon (12:00)

    @property
    def jyear(self):
        """Returns the jyear of the epoch at noon (before the night)."""

        from astropy.time import Time
        from datetime import datetime, date, time

        return Time(datetime.combine(self.night, time(hour=12))).jyear

    # helper properties that return counts to files of this epoch.

    @property
    def rawcount(self):
        from iop4lib.db import RawFit
        return self.rawfits.count()

    @property
    def rawbiascount(self):
        from iop4lib.db import RawFit
        return self.rawfits.filter(imgtype=IMGTYPES.BIAS).count()
    
    @property
    def rawdarkcount(self):
        from iop4lib.db import RawFit
        return self.rawfits.filter(imgtype=IMGTYPES.DARK).count()
    
    @property
    def rawflatcount(self):
        from iop4lib.db import RawFit
        return self.rawfits.filter(imgtype=IMGTYPES.FLAT).count()
    
    @property
    def rawlightcount(self):
        from iop4lib.db import RawFit
        return self.rawfits.filter(imgtype=IMGTYPES.LIGHT).count()

    @property
    def reducedcount(self):
        from iop4lib.db import ReducedFit
        return ReducedFit.objects.filter(rawfit__epoch=self.id).count()

    @property
    def calibratedcount(self):
        from iop4lib.db import ReducedFit
        return ReducedFit.objects.filter(rawfit__epoch=self.id, flags__has=ReducedFit.FLAGS.BUILT_REDUCED).count()

    # repr and str

    def _repr_html_(self):
        return (f"Epoch(id {self.id}):<br>"
                f" - telescope: {self.telescope}<br>"
                f" - night: {self.night}<br>"
                f" - {self.rawcount} rawfits: {self.rawbiascount} bias, {self.rawflatcount} flat, {self.rawlightcount} light<br>"
                f" - summary status: {self.get_summary_rawfits_status()}")

    def _repr_pretty_(self, p, cycle):
        if cycle:
            p.text(f"{self!r}")
        else:
            with p.group(4, '<Epoch(', ')>'):
                p.text(f"id: {self.id},")
                p.breakable()
                p.text(f"telescope: {self.telescope},")
                p.breakable()
                p.text(f"night: {self.night},")
                p.breakable()
                p.text(f"rawcount: {self.rawcount}, rawbiascount: {self.rawbiascount}, rawflatcount: {self.rawflatcount}, rawlightcount: {self.rawlightcount},")
                p.breakable()
                p.text(f"summary status: {self.get_summary_rawfits_status()}")


    def __repr__(self):
        return f"Epoch.objects.get(id={self.id!r})"
    
    def __str__(self):
        return f"<Epoch {self.id} | {self.epochname}>"

    # creator

    @staticmethod
    def epochname_to_tel_night(epochname : str) -> tuple[str, datetime.date]:
        """Parses an epochname to a telescope and night."""
        
        matches = re.findall(r"([a-zA-Z0-9]+)/([0-9]{2,4}-?[0-9]{2}-?[0-9]{2})$", epochname)

        if len(matches) != 1:
            raise Exception(f"Epoch name {epochname} is not TEL/YYMMDD, TEL/YYYYMMDD or TEL/YYYY-MM-DD.")

        telescope = matches[0][0]
        if Telescope.is_known(telescope):
            telescope = Telescope.by_name(telescope).name
        else:
            raise Exception(f"Telescope {telescope} in epochname {epochname} not known.")

        if len(matches[0][1]) == 6: ## yymmdd
            logger.debug("Epoch name was specified in YYMMDD format, it will automatically casted to 20YYMMDD format.")
            yymmdd = matches[0][1]
            yyyymmdd = f"20{yymmdd}"
        elif len(matches[0][1]) == 8: # yyyymmdd
            yyyymmdd = matches[0][1]
            yymmdd = matches[0][1][:-6]
        elif len(matches[0][1]) == 10: ## yyyy-mm-dd
            yyyymmdd = "".join(matches[0][1].split("-"))
            yymmdd = yyyymmdd[2:]
        

        night = datetime.datetime.strptime(yyyymmdd, "%Y%m%d").date()

        return telescope, night

    @classmethod
    def create(cls, 
                 telescope=None, night=None, epochname=None,
                 auto_link_rawfits=True,
                 check_remote_list=False,
                 force_redownload=False,
                 fallback_to_local=True,
                 auto_merge_to_db=True,
                 rawfits_kwargs=dict()):
        """Create an Epoch object for the given telescope and night, reusing an existing DB entry if it exists.

        Parameters
        ----------
        epochname : str
            TEL/YYYY-MM-DD, TEL/YYYYMMDD or TEL/YYMMDD ot (converted to  TEL/20YYMMDD)

        """

        if telescope is not None and night is not None:
            pass
        elif epochname is not None:
            telescope, night = Epoch.epochname_to_tel_night(epochname)
        else:
            raise ValueError("Either telescope and night or epochname must be specified.")
        
        # create entry

        if (epoch := Epoch.objects.filter(telescope=telescope, night=night).first()) is  None:
            logger.debug(f"Creating DB entry for {telescope}/{night}.")
            epoch = cls(telescope=telescope, night=night)
            epoch.save()
        else:
            logger.debug(f"Epoch for {telescope}/{night} already exists in the DB, it will be used instead.")

        # clear flags on creation
        #epoch.clear_flags()

        # instance only attributes
        epoch.auto_link_rawfits = auto_link_rawfits
        epoch.fallback_to_local = fallback_to_local
        epoch.check_remote_list = check_remote_list
        epoch.force_redownload = force_redownload
        epoch.rawfits_kwargs = rawfits_kwargs
        epoch.auto_merge_to_db = auto_merge_to_db

        if auto_merge_to_db:
            epoch.save()
        
        if auto_link_rawfits:
            epoch.link_rawfits()

        return epoch

    def __init__(self, *args, **kwargs):
        """Provides some defaults for attributes that are not stores in DB.

        Should be the same as default kwargs in .create() for consistency.
        """
        super().__init__(*args, **kwargs)
        self.auto_link_rawfits = True
        self.check_remote_list = False
        self.force_redownload = False
        self.fallback_to_local = True
        self.auto_merge_to_db = True
        self.rawfits_kwargs = dict()
      
    # methods

    def list_remote_raw_fnames(self):
        """Checks remote file list and builds self.rawfits from them.
        """
        return Telescope.by_name(self.telescope).list_remote_raw_fnames(self)

            
    def link_rawfits(self):
        """ Links rawfits to this epoch. """

        from iop4lib.db import RawFit

        self.unset_flag(Epoch.FLAGS.LISTED)

        if not os.path.isdir(self.rawfitsdir):
            logger.debug(f"Dir {self.rawfitsdir} does not exist, creating it.")
            os.makedirs(self.rawfitsdir) 

        local_file_L = os.listdir(self.rawfitsdir)

        if self.check_remote_list:
            logger.info(f"{self}: fetching file list for epoch.")
            try:
                self.fnameL = self.list_remote_raw_fnames()
                logger.info(f"{self}: listed {len(self.fnameL)} files in remote.")
                self.unset_flag(Epoch.FLAGS.ERROR_LISTING_REMOTE)
                self.set_flag(Epoch.FLAGS.LISTED)
            except Exception as e:
                logger.warning(f"{self}: error listing remote dir, maybe it is an old epoch?: {e}.")
                self.set_flag(Epoch.FLAGS.ERROR_LISTING_REMOTE)


        if (self.has_flag(Epoch.FLAGS.ERROR_LISTING_REMOTE) or not self.check_remote_list) and self.fallback_to_local:
            if len(local_file_L) == 0:
                logger.error(f"{self}: used fallback to local, but there was an error listing local dir (0 files found)")
                self.set_flag(Epoch.FLAGS.ERROR_LISTING_LOCAL)
            else:
                self.fnameL = local_file_L
                self.unset_flag(Epoch.FLAGS.ERROR_LISTING_LOCAL)
                self.set_flag(Epoch.FLAGS.LISTED)

        if self.has_flag(Epoch.FLAGS.LISTED):

            # we could allow download on create but better to do it on bulk, it is more efficient (less logins). Also CAHA will block you if you login 2 fast 2 furious.

            rawfits = [RawFit.create(epoch=self, filename=filename, auto_procure_local=False, **self.rawfits_kwargs) for filename in self.fnameL]

            missing_files = set(self.fnameL) - set(local_file_L)

            if self.force_redownload:
                Telescope.by_name(self.telescope).download_rawfits(rawfits)
            elif len(missing_files)>0:
                Telescope.by_name(self.telescope).download_rawfits([rawfit for rawfit in rawfits if rawfit.filename in missing_files])

            for rawfit in rawfits:
                rawfit.procure_local_file()

        else:
            logger.error("Epoch not listed, not linking rawfits.")

        if self.auto_merge_to_db:
            self.save()



    def get_summary_rawfits_status(self):
        """
            Returns list of flags present in the rawfits of this epoch.
        """
        from iop4lib.db import RawFit

        flag_vals_present = self.rawfits.all().values_list('flags', flat=True).distinct()
        flag_label_set = set(itertools.chain.from_iterable([RawFit.FLAGS.get_labels(val) for val in flag_vals_present]))
        return ", ".join(flag_label_set)



    def build_masters(self, model, force_rebuild=False):
        import itertools

        # define keywords to be used to build the master
        kw_L = model.margs_kwL

        # for each keyword, get the set of values in the rawfits of this epoch
        kw_set_D = {kw:set(self.rawfits.filter(imgtype=model.imgtype).values_list(kw, flat=True).distinct()) for kw in kw_L}

        # create a list of dictionaries with all the combinations of values for each keyword
        margs_L = [dict(zip(kw_L, prod)) for prod in itertools.product(*kw_set_D.values())]

        if len(margs_L) == 0:
            logger.error(f"No {model._meta.verbose_name} will be built for this epoch since there are no files for it.")
        
        # create master 

        for margs in margs_L:
            try:
                margs['epoch'] = self
                logger.debug(f"{margs=}")
                if self.rawfits.filter(imgtype=model.imgtype, **margs).count() > 0:
                    logger.info(f"Building {model._meta.verbose_name} for {model.margs2str(margs)}.")
                    model.create(**margs, force_rebuild=force_rebuild)
                else:
                    logger.debug(f"No {model._meta.verbose_name} will be built for this margs since there are no files for it.")
            except Exception as e:
                logger.error(f"Error building {model._meta.verbose_name} for {self.epochname} with args {margs}: {e}.")
                self.set_flag(Epoch.FLAGS.ERROR)
            
        if self.auto_merge_to_db:
            self.save() 

    def build_master_biases(self, **kwargs):
        from iop4lib.db import MasterBias
        return self.build_masters(MasterBias, **kwargs)
    
    def build_master_flats(self, **kwargs):
        from iop4lib.db import MasterFlat
        return self.build_masters(MasterFlat, **kwargs)
    
    def build_master_darks(self, **kwargs):
        from iop4lib.db import MasterDark
        return self.build_masters(MasterDark, **kwargs)

    def reduce(self, force_rebuild=False):
        """ Reduces all (LIGHT) rawfits of this epoch. 
        
        If force_rebuild is False, only rawfits that have not been reduced yet (that do not have 
        the BUILT_REDUCED flag) will be reduced, else all rawfits of this epoch of type LIGHT be 
        reduced.

        If iop4conf.nthreads > 1, the reduction will be done in parallel processes.
        """

        Epoch.reduce_rawfits(self.rawfits.filter(imgtype=IMGTYPES.LIGHT), force_rebuild=force_rebuild, epoch=self)
        
        if self.auto_merge_to_db:
            self.save()


    @staticmethod
    def reduce_rawfits(rawfits, force_rebuild=False, epoch=None):
        """ Bulk reduces a list of RawFit in a multiprocessing pool. `rawfits` can be an iterable such as a QuerySet. 
        
        If force_rebuild is False, only rawfits that have not been reduced yet (that do not have 
        the BUILT_REDUCED flag) will be reduced, else all rawfits in the list will be reduced.

        If iop4conf.nthreads > 1, the reduction will be done in parallel processes.

        Parameters
        ----------
        rawfits : iterable
            An iterable of RawFit objects to be reduced.

        Other Parameters
        ----------------
        epoch : Epoch, optional
            If provided, it is used only to print the epoch in the log messages of the main thread.
                
        """

        from iop4lib.db import ReducedFit

        reduced_L = [ReducedFit.create(rawfit=rf, auto_build=False, force_rebuild=False) for rf in rawfits]

        if not force_rebuild:
            reduced_L = [redf for redf in reduced_L if not redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED)]
            
        Epoch.reduce_reducedfits(reduced_L, epoch=epoch)

    
    @staticmethod
    def reduce_reducedfits(reduced_L, epoch=None):
        """ Bulk reduces a list of ReducedFit in a multiprocessing pool. 
        
        If iop4conf.nthreads > 1, the reduction will be done in parallel processes.

        Parameters
        ----------
        reduced_L : list
            A list of ReducedFit objects to be reduced. The .build_file() method of each object is always called, 
            independently of the value of the BUILT_REDUCED flag.

        Other Parameters
        ----------------
        epoch : Epoch, optional
            If provided, it is used only to print the epoch in the log messages of the main thread.

        """

        if len(reduced_L) > 0:
            if iop4conf.nthreads > 1:
                epoch_bulkreduce_multiprocesing(reduced_L, epoch=epoch)
            else:
                epoch_bulkreduce_onebyone(reduced_L, epoch=epoch)
        else:
            logger.info("No files to build.")


    @classmethod
    def by_epochname(cls, epochname):
        telescope, night = cls.epochname_to_tel_night(epochname)
        return cls.objects.filter(telescope=telescope, night=night).get()




    # REDUCTION METHODS



    def compute_relative_photometry(self, redf_qs=None):
        """ Computes the relative photometry results for this epoch.

        Parameters
        ----------
        redf_qs : QuerySet, optional
            If provided, the relative photometry will be computed only for the ReducedFit objects in the QuerySet.
            If not provided, the relative photometry will be computed for all ReducedFit objects in the epoch.
        """

        from .reducedfit import ReducedFit

        if redf_qs is None:
            redf_qs = ReducedFit.objects.filter(epoch=self, obsmode=OBSMODES.PHOTOMETRY, flags__has=ReducedFit.FLAGS.BUILT_REDUCED).order_by('juliandate').all()
        else:
            redf_qs = redf_qs.filter(epoch=self, obsmode=OBSMODES.PHOTOMETRY, flags__has=ReducedFit.FLAGS.BUILT_REDUCED).order_by('juliandate').all()

        for redf in redf_qs:
            redf.compute_relative_photometry()



    def make_polarimetry_groups(self, redf_qs=None):
        """
        To reduce the polarimetry data, we need to group the reduced fits corresponding to rotated polarization angles.

        To group the reduced fits, we make sure they are close (<10min from the julian date field), and that they correspond to the same object, band and exposure time.

        To make sure the correspond to the make object we make use of the OBJECT keyword in the FITS header; in OSN these are the object names (not necessarily the name we use in IOP4 catalog)
        and in CAHA these are the object name + the angle of the polarizer (e.g. "1652+398 0.0 deg"). So we just split the string by ' ' and use the first element.
        TODO: might be better to use the sources_in_field but right now the catalog is pretty incomplete.
        """

        logger.debug(f"{self}: grouping observations for polarimetry...")

        from .reducedfit import ReducedFit

        if redf_qs is None:
            redf_qs = ReducedFit.objects.filter(epoch=self, obsmode=OBSMODES.POLARIMETRY, flags__has=ReducedFit.FLAGS.BUILT_REDUCED).order_by('juliandate', 'filename').all()
        else:
            redf_qs = redf_qs.filter(epoch=self, obsmode=OBSMODES.POLARIMETRY, flags__has=ReducedFit.FLAGS.BUILT_REDUCED).order_by('juliandate', 'filename').all()

        # it should not be necessary to sort by filename, but there have been some ERRORS 
        # where a few files had the same juliandate. This tries to work around that. 
        # With it, if two files have the same juliandate, the one with the lower filename will 
        # be first (e.g. Mrk421_0001R.fit will be before Mrk421_0002R.fit)
        # Keep an eye on it. It should not happen, and it will affect the time of the results.
        # TODO: check why this happened and maybe remove the 'filename' in order_by if it is not necessary anymore.

        # Create a groups with the same keys (object, band, exptime)

        groups_D = dict()

        for i, redf in enumerate(redf_qs):
            
            keys = (
                    ('kwobj', redf.rawfit.header['OBJECT'].split(" ")[0]), 
                    ('instrument', redf.instrument),
                    ('band', redf.band), 
                    ('exptime', redf.exptime)
            )
            
            if keys not in groups_D.keys():
                groups_D[keys] = list()
                
            groups_D[keys].append(redf)

        # output some debug info about the groups made
        for key_T, redf_L in groups_D.items():
            key_D = dict(key_T)
            rotangles_S = sorted(set([redf.rotangle for redf in redf_L]))
            logger.debug(f"{len(redf_L)=}; {key_D.values()}, {rotangles_S}")
    
        # split the groups into subgroups such that every subgroup has at most 4 elements and all rotangles are present in the subgroup

        split_groups_keys = list()
        split_groups = list()

        for key_T, redf_L in groups_D.items():
            key_D = dict(key_T)

            rotangles_S = sorted(set([redf.rotangle for redf in redf_L])) # rotangles available in the redf_L
            
            split_rotangle_D = {rotangle:[redf for redf in redf_L if redf.rotangle==rotangle] for rotangle in rotangles_S}  # to access the redfs in the redfL by rotangle
            
            while any([len(split_rotangle_D[rotangle])>0 for rotangle in rotangles_S]): # while there are redfs for some rotangle, create groups by popping one of each rotangle
                split_groups.append([split_rotangle_D[rotangle].pop(0) for rotangle in rotangles_S if len(split_rotangle_D[rotangle]) > 0])
                split_groups_keys.append(key_D)

        # sort the groups by min(juliandate)
         
        t1_L = [min([redf.juliandate for redf in redf_L]) for redf_L in split_groups]

        split_groups_keys = [x[1] for x in sorted(zip(t1_L, split_groups_keys), key=lambda x: x[0])]
        split_groups = [x[1] for x in sorted(zip(t1_L, split_groups), key=lambda x: x[0])]

        # some warning/debug info about the final sorted groups:

        # all groups should have different min(juliandate), otherwise there is a problem with the data
        if len(t1_L) != len(set(t1_L)):
            logger.warning(f"{self}: error grouping observations for polarimetry: some groups have the same min(juliandate).")

        if iop4conf.log_level == logging.DEBUG: 
            # so we dont lose time if not in debug mode
            for key_D, redf_L in zip(split_groups_keys, split_groups):
                
                t1 = Time(min([redf.juliandate for redf in redf_L]), format="jd").datetime.strftime("%H:%M:%S")
                t2 = Time(max([redf.juliandate for redf in redf_L]), format="jd").datetime.strftime("%H:%M:%S")
                
                logging.debug(f"Group {len(redf_L)=}; {key_D.values()}, {set([redf.rotangle for redf in redf_L])}   ({t1}, {t2})")
                
                for redf in redf_L:
                    logging.debug(f"     -> {redf.rotangle}: {Time(redf.juliandate, format='jd').datetime.strftime('%H:%M:%S')}")

        # return the groups and their keys:

        return split_groups, split_groups_keys


    def compute_relative_polarimetry(self, redf_qs=None):
        """Computes the relative polarimetry results for this epoch.

        Parameters
        ----------
        redf_qs : QuerySet, optional
            If provided, the relative polarimetry will be computed only for the ReducedFit objects in the QuerySet.
            If not provided, the relative polarimetry will be computed for all ReducedFit objects in the epoch.
        """

        from .reducedfit import ReducedFit

        clusters_L, groupkeys_L = self.make_polarimetry_groups(redf_qs=redf_qs)

        logger.info(f"{self}: computing relative polarimetry over {len(groupkeys_L)} polarimetry groups.")
        logger.debug(f"{self}: {groupkeys_L=}")

        f = lambda x: Instrument.by_name(x[1]['instrument']).compute_relative_polarimetry(x[0])

        if iop4conf.nthreads > 1:
            # parallel
            from iop4lib.utils.parallel import parallel_relative_polarimetry
            parallel_relative_polarimetry(groupkeys_L, clusters_L)
        else: 
            # one by one
            for i, (group, keys) in enumerate(zip(clusters_L, groupkeys_L)):
                try:
                    Instrument.by_name(keys['instrument']).compute_relative_polarimetry(group)
                except Exception as e:
                    logger.error(f"{self}: error computing relative polarimetry for group n {i} {keys}: {e}")
                finally:
                    logger.info(f"{self}: computed relative polarimetry for group n {i} {keys}.")







# BULK REDUCTION ONE BY ONE

def epoch_bulkreduce_onebyone(reduced_L: Sequence['ReducedFit'], epoch: Epoch = None) -> None:
    """ Reduces a list of ReducedFit instances one by one."""
    logger.info(f"{epoch}: building {len(reduced_L)} reduced files one by one. This may take a while.")
    for i, redf in enumerate(reduced_L):
        logger.info(f"{epoch}: building reduced file {i+1}/{len(reduced_L)}: {redf}.")
        redf.build_file()










import pytest
from pathlib import Path

from .conftest import TEST_CONFIG

# iop4lib config
import iop4lib.config
iop4conf = iop4lib.Config(config_path=TEST_CONFIG)

# other imports
import os
from pytest import approx

# logging
import logging
logger = logging.getLogger(__name__)

# fixtures
from .fixtures import load_test_catalog


@pytest.mark.skipif(os.getenv("CI") != "true", reason="only for actions CI")
@pytest.mark.django_db(transaction=True)
def test_build_single_proc(load_test_catalog):
    """ Test the whole building process of reduced fits in a single process

        This test is not really necessary since single-process reduction is already tested
        in OSN-T090, therefore it is skipped by default. However CI actions at the momemnt 
        will fail without it, because if the astrometry index files are accessed by multiple 
        processes at the same time before they are catched httpdirfs will fail. To run it 
        locally set the environment variable CI=true.
    """

    from iop4lib.db import Epoch, RawFit, ReducedFit
    from iop4lib.enums import IMGTYPES, SRCTYPES

    epochname_L = ["CAHA-T220/2022-08-27", "CAHA-T220/2022-09-18"]

    epoch_L = [Epoch.create(epochname=epochname, check_remote_list=False) for epochname in epochname_L]

    for epoch in epoch_L:
        epoch.build_master_biases()
        epoch.build_master_flats()

    iop4conf.nthreads = 1

    rawfits = RawFit.objects.filter(epoch__in=epoch_L, imgtype=IMGTYPES.LIGHT).all()
    
    Epoch.reduce_rawfits(rawfits)

    assert (ReducedFit.objects.filter(epoch__in=epoch_L).count() == 4)

    for redf in ReducedFit.objects.filter(epoch__in=epoch_L).all():
        assert (redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED))
        assert not (redf.has_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY))



@pytest.mark.django_db(transaction=True)
def test_build_multi_proc_photopol(load_test_catalog):
    """ Test the whole building process of reduced fits through multiprocessing 
    
    Also tests here relative photometry and polarimetry results and their 
    quality (value + uncertainties) (to avoud losing time reducing them
    in another test function).
    """

    from iop4lib.db import Epoch, RawFit, ReducedFit
    from iop4lib.enums import IMGTYPES, SRCTYPES

    epochname_L = ["CAHA-T220/2022-08-27", "CAHA-T220/2022-09-18"]

    epoch_L = [Epoch.create(epochname=epochname, check_remote_list=False) for epochname in epochname_L]

    for epoch in epoch_L:
        epoch.build_master_biases()
        epoch.build_master_flats()

    iop4conf.nthreads = 4

    rawfits = RawFit.objects.filter(epoch__in=epoch_L, imgtype=IMGTYPES.LIGHT).all()
    
    Epoch.reduce_rawfits(rawfits)

    assert (ReducedFit.objects.filter(epoch__in=epoch_L).count() == 4)

    for redf in ReducedFit.objects.filter(epoch__in=epoch_L).all():
        assert (redf.has_flag(ReducedFit.FLAGS.BUILT_REDUCED))
        assert not (redf.has_flag(ReducedFit.FLAGS.ERROR_ASTROMETRY))

    from iop4lib.db import PhotoPolResult, AstroSource

    epoch = Epoch.by_epochname("CAHA-T220/2022-09-18")

    epoch.compute_relative_polarimetry()

    qs_res = PhotoPolResult.objects.filter(epoch=epoch, astrosource__name="2200+420").all()

    # we expect only one photometry result target in this test dataset for this epoch and source
    assert qs_res.exclude(astrosource__srctype=SRCTYPES.CALIBRATOR).count() == 1

    res = qs_res[0]

    # check that the result is correct to 1.5 sigma or 0.02 mag compared to IOP3
    assert res.mag == approx(13.38, abs=max(1.5*res.mag_err, 0.05))
    # check that uncertainty of the result is less than 0.08 mag
    assert res.mag_err < 0.08

    assert res.p == approx(10.9/100, abs=max(1.5*res.p_err, 1.0/100))
    assert res.p_err < 0.5

    assert res.chi == approx(25.2, abs=max(1.5*res.chi_err, 1.0))
    assert res.chi_err < 1.0



# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)

# django imports
from django.urls import reverse
from django.shortcuts import redirect
from django.contrib.auth import authenticate, login, logout

# iop4lib imports

# other imports

#logging
import logging
logger = logging.getLogger(__name__)



def login_view(request):
    if request.method == "POST":
        username = request.POST["username"]
        password = request.POST["password"]
        user = authenticate(request, username=username, password=password)
        if user is not None:
            login(request, user)
            return redirect(reverse('iop4api:index', args=[['explore', 'catalog']]))
    
    return redirect("{}?login_failed=1".format(reverse('iop4api:index', args=[["login",]])))




def logout_view(request):
    logout(request)
    return redirect('iop4api:index') 




from django.urls import path, include
from django.conf import settings
from django.conf.urls.static import static

from . import views

app_name = "iop4api"

class TabsConverter:
    regex = '.*'

    def to_python(self, value):
        return value.split('/')

    def to_url(self, value):
        return '/'.join(value)
    
from django.urls import register_converter
register_converter(TabsConverter, 'tabs')

urlpatterns = []

if settings.DEBUG:
    # local docs only accesible in debug server, in production you should 
    # configure it yourself (if you want it to be public, etc)
    urlpatterns += [path(r'docs/', views.docs, name='docs')]
    urlpatterns += [path(r'docs/<path:file_path>', views.docs)]


urlpatterns = urlpatterns + [
    path('', views.index, name='index'),
    path('api/login/', views.login_view, name='login_view'),
    path('api/logout/', views.logout_view, name='logout_view'),
    path('api/plot/', views.plot, name='plot'),
    path('api/data/', views.data, name='data'),
    path('api/catalog/', views.catalog, name='catalog'),
    path('api/log/', views.log, name='log'),
    path('api/flag/', views.flag, name='flag'),
    path('<tabs:tabs>/', views.index, name='index'),
] + static(settings.STATIC_URL, document_root=settings.STATIC_ROOT)



from django.contrib import admin

from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render

# other imports
from iop4api.filters import *
from iop4api.models import *
from iop4api import views

# other imports
from iop4lib.enums import *
from astropy.time import Time

class AdminAperPhotResult(admin.ModelAdmin):
    model = AperPhotResult
    list_display = ['id', 'get_telescope', 'get_instrument', 'get_datetime', 'get_src_name', 'get_src_type', 'get_fwhm', 'get_aperpix', 'get_reducedfit', 'get_obsmode', 'pairs', 'get_rotangle', 'get_src_type', 'get_flux_counts', 'get_flux_counts_err', 'get_bkg_flux_counts', 'get_bkg_flux_counts_err', 'modified']
    readonly_fields = [field.name for field in AperPhotResult._meta.fields]
    search_fields = ['id', 'reducedfit__instrument', 'astrosource__name', 'astrosource__srctype', 'reducedfit__id']
    list_filter = ['reducedfit__instrument', 'astrosource__srctype', 'reducedfit__epoch__telescope', 'reducedfit__obsmode']


    
    @admin.display(description="TELESCOPE")
    def get_telescope(self, obj):
        return obj.reducedfit.epoch.telescope

    @admin.display(description="INSTRUMENT")
    def get_instrument(self, obj):
        return obj.reducedfit.instrument

    @admin.display(description="DATETIME")
    def get_datetime(self, obj):
        return Time(obj.reducedfit.juliandate, format='jd').strftime('%Y-%m-%d %H:%M:%S')
    
    @admin.display(description="SRCNAME")
    def get_src_name(self, obj):
        return obj.astrosource.name
    
    @admin.display(description="SRCTYPE")
    def get_src_type(self, obj):
        return obj.astrosource.srctype
    
    @admin.display(description="ReducedFit")
    def get_reducedfit(self, obj):
        self.allow_tags = True
        url = reverse('iop4admin:%s_%s_changelist' % (ReducedFit._meta.app_label, ReducedFit._meta.model_name)) + f"?id={obj.reducedfit.id}"
        return mark_safe(rf'<a href="{url}">{obj.reducedfit.id}</a>')    
    
    @admin.display(description="OBSMODE")
    def get_obsmode(self, obj):
        return obj.reducedfit.obsmode
    
    @admin.display(description="ROTANGLE")
    def get_rotangle(self, obj):
        return obj.reducedfit.rotangle
    
    @admin.display(description="fwhm")
    def get_fwhm(self, obj):
        if obj.fwhm is None:
            return "-"
        return f"{obj.fwhm:.1f}"
    
    @admin.display(description="aperpix")
    def get_aperpix(self, obj):
        if obj.aperpix is None:
            return "-"
        return f"{obj.aperpix:.1f}"
    
    @admin.display(description="flux_counts")
    def get_flux_counts(self, obj):
        if obj.flux_counts is None:
            return "-"
        else:
            return f"{obj.flux_counts:.1f}"
    
    @admin.display(description="flux_counts_err")
    def get_flux_counts_err(self, obj):
        if obj.flux_counts_err is None:
            return "-"
        else:
            return f"{obj.flux_counts_err:.1f}"
    
    @admin.display(description="bkg_flux_counts")
    def get_bkg_flux_counts(self, obj):
        if obj.bkg_flux_counts is None:
            return "-"
        else:
            return f"{obj.bkg_flux_counts:.1f}"
    
    @admin.display(description="bkg_flux_counts_err")
    def get_bkg_flux_counts_err(self, obj):
        if obj.bkg_flux_counts_err is None:
            return "-"
        else:
            return f"{obj.bkg_flux_counts_err:.1f}"
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""This script will create a iop4testdata.tar.gz file in your home directory.

Use it to build a test dataset from your reduced and tested results. The .tar.gz 
will contain all the raw data needed to reduce the given photo-polarimetry results,
including raw light, bias, darks, flats and a test catalog file.

Edit the pk_L list to add the pk of the results you want to include in the test dataset.

Edit the raw_fileloc_L list to add the paths of additional files you want to include in the test dataset.

"""

import iop4lib.config
iop4conf = iop4lib.Config(config_db=True)

from iop4lib.db import Epoch, RawFit, ReducedFit, PhotoPolResult, MasterBias, MasterFlat, AperPhotResult, AstroSource
from iop4lib.telescopes import Telescope

import os
import subprocess
import shutil
import hashlib
from pathlib import Path


######################
### Configuration ####
######################


clean_workdir = False

Nmax_raw_per_master = 1

# only Nmax_raw_per_master will be used for these files

fileloc_L = [
    # some dipol files to test calibration
    "OSN-T090/2023-11-06/BLLac_IAR-0001R.fit", # DIPIL photometry astrocalibration / shotgun
    "OSN-T090/2023-10-11/OJ248_R_IAR-0111.fts", # DIPOL polarimetry astrocalibration / catalog matching in a blazar
    "OSN-T090/2023-10-25/HD204827_R_IAR-0384.fts", # DIPOL polarimetry astrocalibration / target E, O in a star
    # DIPOL polarimetry astrocalibration / quad matching in a blazar
    "OSN-T090/2023-11-06/BLLac_IAR-0001R.fit", # the photometry file
    "OSN-T090/2023-11-06/BLLAC_R_IAR-0760.fts", # the polarimetry file
]

# for this, Nmax_raw_per_master will be ignored (to keep the result as close to the real one as possible)

pk_L = [
            # 48572, # OSN-T090 Andor Polarimetry 2022-09-08 23:03:14 2200+420 reduced fits 6565, 6566, 6567, 6568  mag R 13.328 \pm 0.164 p 12.48 % \pm 0.35 % chi 15.38 \pm 0.79, iop3 mag 13.38, p 14.0, chi 14.7
            # 34354, # OSN-T090 Andor Photometry 2022-09-18 23:05:05 2200+420 reduced fits 39071 mag R 13.292 \pm 0.034, iop3 mag 13.35
            # 64092, # CAHA-T220 CAFOS Polarimetry 2022-09-18 23:02:53 2200+420 reduced fits 40827, 40828, 40829, 40830 mag R 13.369 \pm 0.036 p 11.18 % \pm 0.12 % chi 25.45 \pm 0.30, iop3 mag 13.38, p 10.9, chi 25.2
            # ?????, # OSN-T090 DIPOL Polarimetry 2023-10-11 3C345 [mjd 60228.82199 datetime 2023/10/11 19:43	p 29.238 ± 0.000 chi 38.379 ± 0.392]
        ]

output_file = Path("~/iop4testdata.tar.gz").expanduser()
workdir = Path("~/iop4testdata").expanduser()

# remove workdir if configured, and create it again if necessary

if clean_workdir and os.path.exists(workdir):
    shutil.rmtree(workdir)

os.makedirs(workdir, exist_ok=True)


###############################################
### Files for the PhotoPolResults pk_L list ###
###############################################


print("Copying files for the PhotoPolResults pk_L list")

# get the list of files needed and copy them to the workdir

rawfitL = list()

for pk in pk_L:
    res = PhotoPolResult.objects.get(id=pk)

    if res.instrument == "DIPOL":
        Nmax = 1 # too many files otherwise
    else:
        Nmax = None

    for redf in res.reducedfits.all():
        # append all rawfits needed for this reducedfit
        rawfitL.append(redf.rawfit)

        # the bias for its masterbias
        for bias in redf.masterbias.rawfits.all()[:None]:
            rawfitL.append(bias)

        # the darks for its masterdark
        if redf.masterdark is not None:
            for dark in redf.masterdark.rawfits.all()[:None]:
                rawfitL.append(dark)

        # the flats for its masterflat
        for flat in redf.masterflat.rawfits.all()[:None]:
            rawfitL.append(flat)

        # the bias for the masterbias of its masterflat
        for bias in redf.masterflat.masterbias.rawfits.all()[:None]:
            rawfitL.append(bias)

        # the bias for the masterbias of its masterdark
        if redf.masterdark is not None:
            for bias in redf.masterdark.masterbias.rawfits.all()[:None]:
                rawfitL.append(bias)

        # the bias and the darks for the masterdark of its masterflat
        if redf.masterflat.masterdark is not None:
            for dark in redf.masterflat.masterdark.rawfits.all()[:None]:
                rawfitL.append(dark)


files_to_download = set([rawfit for rawfit in rawfitL if not os.path.exists(rawfit.filepath)])

if len(files_to_download) > 0:
    # download the files if needed
    Telescope.get_by_name(rawfitL[0].epoch.telescope).download_rawfits(files_to_download)

for rawfit in rawfitL:

    print(f"  cp {rawfit.fileloc}")

    relative_path = Path(rawfit.filepath).relative_to(iop4conf.datadir)
    
    dest = workdir / relative_path

    if not os.path.exists(os.path.dirname(dest)):
        os.makedirs(os.path.dirname(dest))

    if not os.path.exists(dest):
        shutil.copy(rawfit.filepath, dest)



###############################
### create the catalog file ###
###############################

print("Creating the catalog file")

astrosources_ids_L = list()

for res in PhotoPolResult.objects.filter(id__in=pk_L):
    astrosources_ids_L.append(res.astrosource.id)

    for calibrator in AstroSource.objects.filter(calibrates=res.astrosource):
        astrosources_ids_L.append(calibrator.id)

from django.core.serializers import serialize
with open(workdir / "testcatalog.yaml", "a") as f:
    f.write(serialize("yaml", AstroSource.objects.filter(id__in=set(astrosources_ids_L)), use_natural_foreign_keys=True, use_natural_primary_keys=True))
    f.write('\n')



#############################
### copy additional files ###
#############################

print("Copying additional files")

rawfitL = list()

for fileloc in fileloc_L:

    redf = ReducedFit.by_fileloc(fileloc)

    # append all rawfits needed for this reducedfit
    rawfitL.append(redf.rawfit)

    # the bias for its masterbias
    for bias in redf.masterbias.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(bias)

    # the darks for its masterdark
    if redf.masterdark is not None:
        for dark in redf.masterdark.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(dark)

    # the flats for its masterflat
    for flat in redf.masterflat.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(flat)

    # the bias for the masterbias of its masterflat
    for bias in redf.masterflat.masterbias.rawfits.all()[:Nmax_raw_per_master]:
        rawfitL.append(bias)

    # the bias for the masterbias of its masterdark
    if redf.masterdark is not None:
        for bias in redf.masterdark.masterbias.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(bias)

    # the bias and the darks for the masterdark of its masterflat
    if redf.masterflat.masterdark is not None:
        for dark in redf.masterflat.masterdark.rawfits.all()[:Nmax_raw_per_master]:
            rawfitL.append(dark)


for rawfit in rawfitL:

    print(f"  cp {rawfit.fileloc}")

    relative_path = Path(rawfit.filepath).relative_to(iop4conf.datadir)
    
    dest = workdir / relative_path

    if not os.path.exists(os.path.dirname(dest)):
        os.makedirs(os.path.dirname(dest))

    if not os.path.exists(dest):
        shutil.copy(rawfit.filepath, dest)
        

###########################
### Create .tar.gz file ###
###########################

print("Creating .tar.gz file")

# in mac-os, additional ._* files are created to save xattrs, avoid it setting COPYFILE_DISABLE
os.environ["COPYFILE_DISABLE"] = "true"

result = subprocess.run(["tar", "--no-xattrs", "--exclude='.*'", "-czf", str(output_file), "-C", str(workdir.parent), str(workdir.name)])

if result.returncode != 0:
    print("Error creating the .tar.gz file")
    exit(1)

print("Created {}".format(output_file))

# read it and give the md5 sum

with open(output_file, "rb") as f:
    md5 = hashlib.md5(f.read()).hexdigest()
    print("{} {}".format(output_file.name, md5))

exit(0)


from django.contrib import admin

from django.utils.html import format_html
from django.utils.safestring import mark_safe
from django.urls import path, reverse 
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.shortcuts import render

from iop4api.filters import *
from iop4api.models import *

# other imports
from iop4lib.enums import *
from astropy.time import Time

class AdminPhotoPolResult(admin.ModelAdmin):
    model = PhotoPolResult
    list_display = ['id', 'get_telescope', 'get_juliandate', 'get_datetime', 'get_src_name', 'get_src_type', 'get_reducedfits', 'obsmode', 'band', 'exptime', 'get_mag', 'get_mag_err', 'get_p', 'get_p_err', 'get_chi', 'get_chi_err', 'get_aperpix', 'get_aperas', 'get_flags', 'modified']
    readonly_fields = [field.name for field in PhotoPolResult._meta.fields]
    search_fields = ['id', 'astrosource__name', 'astrosource__srctype', 'epoch__night']
    ordering = ['-juliandate']
    list_filter = ['astrosource__srctype', 'epoch__telescope', 'obsmode']


    
    @admin.display(description="TELESCOPE")
    def get_telescope(self, obj):
        return obj.epoch.telescope

    @admin.display(description="DATETIME", ordering='-juliandate')
    def get_datetime(self, obj):
        return Time(obj.juliandate, format='jd').strftime('%Y-%m-%d %H:%M:%S')
    
    @admin.display(description="SRCNAME")
    def get_src_name(self, obj):
        return obj.astrosource.name
    
    @admin.display(description="SRCTYPE")
    def get_src_type(self, obj):
        return obj.astrosource.srctype
    
    @admin.display(description="ReducedFits")
    def get_reducedfits(self, obj):
        self.allow_tags = True
        
        ids_str_L = [str(reducedfit.id) for reducedfit in obj.reducedfits.all()]
        a_href = reverse('iop4admin:%s_%s_changelist' % (ReducedFit._meta.app_label, ReducedFit._meta.model_name)) + "?id__in=%s" % ",".join(ids_str_L)
        a_text = ", ".join(ids_str_L)
        return mark_safe(f'<a href="{a_href}">{a_text}</a>')
    
    @admin.display(description="JD", ordering='-juliandate')
    def get_juliandate(self, obj):
        return f"{obj.juliandate:.6f}"
    
    @admin.display(description="MAG", ordering='-mag')
    def get_mag(self, obj):
        return f"{obj.mag:.3f}" if obj.mag is not None else None
    
    @admin.display(description="MAGERR", ordering='-mag_err')
    def get_mag_err(self, obj):
        return f"{obj.mag_err:.3f}" if obj.mag_err is not None else None
    
    @admin.display(description="P [%]", ordering='-p')
    def get_p(self, obj):
        return f"{100*obj.p:.2f}" if obj.p is not None else None
    
    @admin.display(description="PERR [%]", ordering='-p_err')
    def get_p_err(self, obj):
        return f"{100*obj.p_err:.2f}" if obj.p_err is not None else None
    
    @admin.display(description="CHI [º]", ordering='-chi')
    def get_chi(self, obj):
        return f"{obj.chi:.2f}" if obj.chi is not None else None
    
    @admin.display(description="CHIERR [º]", ordering='-chi_err')
    def get_chi_err(self, obj):
        return f"{obj.chi_err:.2f}" if obj.chi_err is not None else None
    
    @admin.display(description='aperpix', ordering='-aperpix')
    def get_aperpix(self, obj):
        return f"{obj.aperpix:.2f}" if obj.aperpix is not None else None
    
    @admin.display(description='aperas', ordering='-aperas')
    def get_aperas(self, obj):
        return f"{obj.aperas:.2f}" if obj.aperas is not None else None
    
    @admin.display(description='Status')
    def get_flags(self, obj):
        return ", ".join(list(obj.flag_labels))


import iop4lib.config
iop4conf = iop4lib.Config(config_db=False)

import os
import dill
from functools import wraps

import logging
logger = logging.getLogger(__name__)

class FiledProperty():
    """ An instance attributed stored in the .filedpropdir of the instance as a pickle file. """
    
    def __init__(self, fget=None, fset=None, name=None, readonly=False):  
        self.readonly = readonly
        self.fget = fget
        self.fset = fset
        self.name = name

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name

        self.owner = owner

    def setter(self, func):
        prop = type(self)(fget=self.fget, fset=func, name=self.name)
        return prop
        
    def getter(self, func):
        prop = type(self)(fget=func, fset=self.fset, name=self.name)
        return prop

    def __get__(self, obj, objtype=None):
        if obj is None:
            return self
            
        if self.fget is None:
            return self._fget_default(obj)
        else:
            return self.fget(obj)
        
    def __set__(self, obj, value):  
        if obj is None:
            return self
            
        if self.fset is None:
            if self.readonly:
                raise AttributeError("Can not set attribute")
            else:
                return self._fset_default(obj, value)
        else:
            return self.fset(obj, value)

    def _get_prop_path(self, obj):
        if not hasattr(obj, 'filedpropdir'):
            raise AttributeError("Object has no filedpropdir attribute")
        
        if not os.path.isdir(obj.filedpropdir):
            os.makedirs(obj.filedpropdir)

        return os.path.join(obj.filedpropdir, self.name)
    
    def _fget_default(self, obj):
        prop_path = self._get_prop_path(obj)
        if not os.path.isfile(prop_path):
            raise NameError
        with open(prop_path, 'rb') as f:
            return dill.load(f)
        
    def _fset_default(self, obj, value):
        prop_path = self._get_prop_path(obj)
        with open(prop_path, 'wb') as f:
            dill.dump(value, f)
            

def filed_property(name=None, readonly=True):
    
    def decorator(fget):
        if name is None:
            return FiledProperty(name=fget.__name__, readonly=readonly, fget=fget)
        if name is not None:
            return FiledProperty(name=name, readonly=readonly, fget=fget)
        
    return decorator


from django.shortcuts import render

from django.urls import reverse
from django.http import HttpResponse, HttpResponseRedirect, HttpResponseNotFound
from django.contrib.auth.decorators import login_required
from django.template import loader

from django.contrib.admin.views.decorators import staff_member_required

from ..models import *
# from .. import admin


# IOP4API 

from .index import index
from .plot import plot
from .catalog import catalog
from .data import data
from .auth import login_view, logout_view
from .log import log
from .flag import flag
from .docs import docs




# iop4lib config
import iop4lib
iop4conf = iop4lib.Config(config_db=False)


# django imports
from django.http import JsonResponse, HttpResponseBadRequest
from django.contrib.auth.decorators import permission_required

# iop4lib imports
from iop4lib.db import PhotoPolResult, AstroSource
from iop4lib.utils import get_column_values

# other imports
import os
import glob
import numpy as np
import pandas as pd
from astropy.time import Time
from pathlib import Path

#logging
import logging
logger = logging.getLogger(__name__)


@permission_required(["iop4api.view_photopolresult", "iop4api.view_astrosource"])
def plot(request):

    source_name = request.POST.get("source_name", None)
    band = request.POST.get("band", "R")
    date_start = request.POST.get("date_start", None) or None # it might be an empty string, this converts it to None
    date_end = request.POST.get("date_end", None) or None

    enable_crosscheckpts = request.POST.get("enable_crosscheckpts", False)
    enable_full_lc = request.POST.get("enable_full_lc", False)
    enable_errorbars = request.POST.get("enable_errorbars", False)
    use_hostcorrected = request.POST.get("use_hostcorrected", False)

    save_filename = f"IOP4_{source_name}_{band}"

    # comment these lines to allow for empty plot
    if not AstroSource.objects.filter(name=source_name).exists(): 
        return HttpResponseBadRequest(f"Source '{source_name}' does not exist".format(source_name=source_name))
    
    target_src = AstroSource.objects.get(name=source_name)

    logger.debug(f"Serving results for {target_src} band {band} from {date_start} to {date_end}")

    # build plot

    import bokeh.colors
    from bokeh.document import Document
    from bokeh.transform import factor_cmap
    from bokeh.layouts import column, gridplot
    from bokeh.models import CategoricalColorMapper, LinearColorMapper, RangeTool, Range1d, LinearAxis, CustomJS, ColumnDataSource, Whisker, DatetimeAxis, DatetimeTickFormatter, Scatter, Segment, CDSView, GroupFilter, AllIndices, HoverTool, NumeralTickFormatter, BoxZoomTool, DataTable, TableColumn, CategoricalMarkerMapper, CustomJSTransform, CustomJSHover, SaveTool, Toolbar
    from bokeh.plotting import figure, show
    from bokeh.embed import components, json_item
    from bokeh.models import Div, Circle, Column, Row
    from bokeh.plotting import figure
    from bokeh.models import ColumnDataSource, Styles, InlineStyleSheet, GlobalInlineStyleSheet
    from bokeh.transform import factor_cmap, factor_mark, transform
    from bokeh.events import Event, PanStart, PanEnd, Press, PressUp, RangesUpdate
    from bokeh.models.widgets import HTMLTemplateFormatter, NumberFormatter

    lod_threshold = 2000
    lod_factor = 10
    lod_timeout = 500

    # secondary axes values
    @np.vectorize
    def f_x1_to_x2(x1_val):
        return Time(x1_val, format='mjd').datetime
    
    # Get data from DB
    column_names = ['id', 'juliandate', 'band', 'instrument', 'mag', 'mag_err', 'p', 'p_err', 'chi', 'chi_err', 'flags']

    if use_hostcorrected:
        column_names += ['mag_corr', 'mag_corr_err', 'p_corr', 'p_corr_err']

    qs = PhotoPolResult.objects.filter(astrosource__name=source_name, band=band).all()

    if date_start is not None:
        qs = qs.filter(juliandate__gte=Time(date_start).jd)

    if date_end is not None:
        qs = qs.filter(juliandate__lte=Time(date_end).jd)

    # crosscheck data

    df_crosscheck = pd.DataFrame(columns=['juliandate', 'mag', 'dmag', 'p', 'p_err', 'chi', 'chi_err', 'instrument'])

    if enable_crosscheckpts:

        iop3_fpath = str(Path(iop4conf.datadir) / "iop4_crosscheck_data" / "iop3.csv")

        if os.path.exists(iop3_fpath):
            df = pd.read_csv(iop3_fpath)

            df_idx = (df["name_IAU"] == source_name) & (df["filter"] == band)

            if date_start is not None:
                df_idx = df_idx & (df["mjd_obs"] >= Time(date_start).mjd)
            if date_end is not None:
                df_idx = df_idx & (df["mjd_obs"] <= Time(date_end).mjd)

            df = df.loc[df_idx]

            # rename IOP3 dump columns to match PhotoPolResult
            df['juliandate'] = Time(df["mjd_obs"], format="mjd").jd
            df['mag'] = df['Mag']
            df['dmag'] = df['dMag']
            df['chi'] = df['Theta']
            df['chi_err'] = df['dTheta']
            df['p'] = df['P']/100
            df['p_err'] = df['dP']/100
            df['instrument'] = list(map(lambda x: "IOP3-"+x, df_crosscheck["Telescope"]))

            df_crosscheck = pd.concat([df_crosscheck, df], ignore_index=True, join="inner")
        
        dipol_dpath = str(Path(iop4conf.datadir) / "iop4_crosscheck_data" / "dipol")
        
        def get_invariable_str(s):
            return s.replace(' ', '').replace('-','').replace('+','').replace('_','').upper()
    
        for fpath in glob.iglob(f"{dipol_dpath}/*.txt"):

            if get_invariable_str(target_src.name) in get_invariable_str(fpath) or get_invariable_str(target_src.other_name) in get_invariable_str(fpath):
                logger.debug(f"Reading {fpath}")

                df = pd.read_csv(fpath, delim_whitespace=True, header=None, 
                                 names=["mjd", "q", "u", "p", "p_err", "chi", "chi_err"],
                                 usecols=range(7), dtype=float)

                if date_start is not None:
                    df = df.loc[df["mjd"] >= Time(date_start).mjd]
                if date_end is not None:
                    df = df.loc[df["mjd"] <= Time(date_end).mjd]

                df['juliandate'] = Time(df["mjd"], format="mjd").jd
                df['mag'] = np.nan
                df['dmag'] = np.nan
                df['p'] = df['p']/100
                df['p_err'] = df['p_err']/100
                df['instrument'] = np.full(len(df), "VILPPU-DIPOL")

                df_crosscheck = pd.concat([df_crosscheck, df], ignore_index=True, join="inner")

    # choose  the x and y
    if qs.count() > 0:
        vals = get_column_values(qs, column_names)
        if use_hostcorrected:
            # just overwrite the values
            vals['mag'] = vals['mag_corr']
            vals['mag_err'] = vals['mag_corr_err']
            vals['p'] = vals['p_corr']
            vals['p_err'] = vals['p_corr_err']
    else:
        vals = {k:np.array([]) for k in column_names}

    if enable_crosscheckpts and len(df_crosscheck) > 0:
        vals["instrument"] = np.append(vals["instrument"], df_crosscheck['instrument'])
        vals["id"] = np.append(vals["id"], -np.arange(len(df_crosscheck))) ## assign negative ids to crosscheck data
        vals["juliandate"] = np.append(vals["juliandate"], df_crosscheck['juliandate'])
        vals["mag"] = np.append(vals["mag"],  df_crosscheck['mag'])
        vals["mag_err"] = np.append(vals["mag_err"], df_crosscheck['dmag'])
        vals["p"] = np.append(vals["p"], df_crosscheck['p'])
        vals["p_err"] = np.append(vals["p_err"], df_crosscheck['p_err'])
        vals["chi"] = np.append(vals["chi"], df_crosscheck['chi'])
        vals["chi_err"] = np.append(vals["chi_err"], df_crosscheck['chi_err'])
        vals["flags"] = np.append(vals["flags"], np.full(len(df_crosscheck), 0)) # do not show flags for crosscheck data (0 means no flags)

    if len(vals['id']) > 0:
        pks = vals['id']
        x1 = Time(vals['juliandate'], format='jd').mjd
        x2 = f_x1_to_x2(x1)
        y1 = vals['mag']
    else: # to allow empty plot
        vals = {k:np.array([]) for k in column_names}
        pks = np.array([])
        x1 = np.array([])
        x2 = np.array([])
        y1 = np.array([])

    # Instruments and colors (auto)
    instruments_L = sorted(list(set(vals['instrument'])), reverse=True)

    #instrument_color_L = [bokeh.palettes.TolRainbow[max(3,len(instruments_L))][i % len(instruments_L)] for i in range(len(instruments_L))]
    instrument_color_L = [bokeh.palettes.Category10[max(3,len(instruments_L))][i % len(instruments_L)] for i in range(len(instruments_L))]

    rgb_L = [bokeh.colors.RGB.from_hex_string(hexcolor) for hexcolor in instrument_color_L]

    hsl_L_nonselected = [rgb.to_hsl() for rgb in rgb_L]
    hsl_L_selected= [rgb.to_hsl() for rgb in rgb_L]

    for hsl in hsl_L_nonselected:
        # make non-selected colors lighter and more gray
        hsl.l = 0.8
        hsl.s = 0.5

    for hsl in hsl_L_selected:
        hsl.l = 0.5
    
    palette_selected = [bokeh.colors.RGB.from_hsl(hsl) for hsl in hsl_L_selected]
    palette_nonselected = [bokeh.colors.RGB.from_hsl(hsl) for hsl in hsl_L_nonselected]
    
    index_cmap_selected = factor_cmap('instrument', 
                             palette=palette_selected, 
                             factors=instruments_L)
    
    index_cmap_nonselected = factor_cmap('instrument', 
                                   palette=palette_nonselected, 
                                   factors=instruments_L)
    
    # # Flags and markers

    markermap = transform("flags", CustomJSTransform(v_func="""
            let markers = [];
            
            for (let i = 0; i < xs.length; i++) {
                                                     
                let marker = "circle";
                                                                                          
                if (xs[i] & (1 << 1)) { // bad photometry
                    marker = "triangle";
                } 

                if (xs[i] & (1 << 2)) { // bad polarimetry
                    marker = "inverted_triangle";
                }

                if ((xs[i] & (1 << 1)) && (xs[i] & (1 << 2))) { // bad photometry and polarimetry
                    marker = "cross";                                 
                }
                                                                                                                       
                markers.push(marker);
            }   
                                                     
            return markers;"""))


    # Define data sources

    source = ColumnDataSource(data=dict(pk = pks,
                                        instrument = vals['instrument'],
                                        x1 = x1, # mjd
                                        x2 = x2, # datetime
                                        datestr = Time(vals["juliandate"],format="jd").strftime("%Y/%m/%d %H:%M"),
                                        y1 = vals['mag'], 
                                        y1_min = vals['mag']-vals['mag_err'], # if not sent it is computed in JS in check_plot()
                                        y1_max = vals['mag']+vals['mag_err'], # if not sent it is computed in JS in check_plot()
                                        y1_err = vals['mag_err'],
                                        y2 = vals['p'], 
                                        y2_min = vals['p']-vals['p_err'], # if not sent it is computed in JS in check_plot()
                                        y2_max = vals['p']+vals['p_err'], # if not sent it is computed in JS in check_plot()
                                        y2_err = vals['p_err'],
                                        y3 = vals['chi'], 
                                        y3_min = vals['chi']-vals['chi_err'], # if not sent it is computed in JS in check_plot()
                                        y3_max = vals['chi']+vals['chi_err'], # if not sent it is computed in JS in check_plot()
                                        y3_err = vals['chi_err'],
                                        flags = vals['flags'],
                                    ),
                                name="source")

    view = CDSView(filter=AllIndices(), name="plot_view")

    # These will set the correspondence between axes x1 units and axes x2 units 

    if len(x1) >= 2:
        # for the main plot
        x1_lims = min(x1)-0.2*(max(x1)-min(x1)), max(x1)+0.2*(max(x1)-min(x1))
        x2_lims = f_x1_to_x2(x1_lims)

        # and for the subplots
        x1_range = min(x1)-0.05*(max(x1)-min(x1)), max(x1)+0.05*(max(x1)-min(x1))
        x2_range = f_x1_to_x2(x1_range)
    elif len(x1) == 1:
        x1_lims = x1[0]-0.2, x1[0]+0.2
        x2_lims = f_x1_to_x2(x1_lims)

        x1_range = x1[0]-0.05, x1[0]+0.05
        x2_range = f_x1_to_x2(x1_range)
    else:
        x1_lims = np.nan, np.nan
        x2_lims = np.nan, np.nan
        x1_range = np.nan, np.nan
        x2_range = np.nan, np.nan

    # also the freeze y axis range of the main plot

    x1 = np.array(x1, dtype=float)
    y1 = np.array(y1, dtype=float)

    if len(y1) >= 2:
        y1_lims = np.nanmin(y1)-0.05*(np.nanmax(y1)-np.nanmin(y1)), np.nanmax(y1)+0.05*(np.nanmax(y1)-np.nanmin(y1))
    elif len(y1) == 1:
        y1_lims = y1[0]-0.05, y1[0]+0.05
    else:
        y1_lims = np.nan, np.nan

    # Create RangeTool and Selected_range
    selected_range = Range1d(*x1_range)

    if enable_full_lc:
        range_tool = RangeTool(x_range=selected_range)
        range_tool.overlay.fill_color = "navy"
        range_tool.overlay.fill_alpha = 0.2

    # Create a hover tool with custom JS callback

    hover_formatter = CustomJSHover(args=dict(source=source), 
                                    code="""
                                        let idx = special_vars.index;

                                        const [name, s1, s2] = format.split(',');
                                        const y = source.data[s1][idx];
                                        const dy = source.data[s2][idx];

                                        if (name == "mag") {
                                            return parseFloat(y).toFixed(2) + " ± " + parseFloat(dy).toFixed(2);
                                        } else if ((name == "p")) {
                                            return parseFloat(100*y).toFixed(2) + " ± " + parseFloat(100*dy).toFixed(2) + " %";
                                        } else if ((name == "chi")) {
                                            return parseFloat(y).toFixed(2) + " ± " + parseFloat(dy).toFixed(2) + " º";
                                        }

                                        return "error in hover formatter";
                                        """)

    hover_tool = HoverTool(renderers=[], formatters={"@pk":hover_formatter})

    hover_tool.callback = CustomJS(args = dict(source = source, hover = hover_tool), code = '''
            if (cb_data.index.indices.length > 0) { 
                let index = cb_data.index.indices[0];
                              
                let mag = source.data.y1[index];
                let p = source.data.y2[index];
                let chi = source.data.y3[index];
                let flags = source.data.flags[index];

                let tooltips = [["id", "@pk"], ["instrument", "@instrument"], ["date", "@datestr"]]
                              
                if (isFinite(mag)) {
                    //tooltips.push(["mag", "@y1"]);
                    tooltips.push(["mag", "@pk{mag,y1,y1_err}"]);
                }
                             
                if (isFinite(p)) {
                    //tooltips.push(["mag", "@y2{0.0 %}"]);
                    tooltips.push(["p", "@pk{p,y2,y2_err}"]);                                       
                } 
                              
                if (isFinite(chi)) {
                    //tooltips.push(["mag", "@y2"]);
                    tooltips.push(["chi", "@pk{chi,y3,y3_err}"]);
                }
                    
                tooltips.push(["flags", flags_to_str(flags)]);
                              
                hover.tooltips = tooltips
                
            } ''')
    
    # other tools

    save_tool = SaveTool(filename=save_filename) # customize filename of saved plot
    # tools = ["fullscreen", "reset", "pan", "wheel_zoom", "box_select", "tap", "lasso_select", save_tool, "auto_box_zoom"]
    box_zoom_tool = BoxZoomTool(dimensions="auto") # added separately in each figure to make it the default active tool
    tools = ["fullscreen", "reset", "pan", "wheel_zoom", "box_select", "tap", "lasso_select", save_tool]

    if enable_full_lc:
        # Create the main plot with range slider and secondary x-axis, fixed styles of markers

        # alternatively: # figure(sizing_mode="stretch", width=800, height=200,
        p_main = figure(x_range=x1_lims, tools="", lod_threshold=lod_threshold, lod_factor=lod_factor, lod_timeout=lod_timeout, output_backend="webgl")  
        p_main.circle(x='x1', y='y1', source=source, view=view,
                    size=3, 
                    line_color=bokeh.colors.named.navy, 
                    fill_color=bokeh.colors.named.navy, 
                    selection_line_color=bokeh.colors.named.navy.darken(0.1), 
                    nonselection_line_color=bokeh.colors.named.navy.lighten(0.55), 
                    selection_fill_color=bokeh.colors.named.navy.darken(0.1), 
                    nonselection_fill_color=bokeh.colors.named.navy.lighten(0.55), 
                    selection_fill_alpha=0.5, 
                    nonselection_fill_alpha=0.5, 
                    selection_line_alpha=0.5, 
                    nonselection_line_alpha=0.5)

        p_main.y_range = Range1d(*y1_lims[::-1])

        p_main.extra_x_ranges["secondary"] = Range1d(*x2_lims)
        p_main_ax_x2 = DatetimeAxis(x_range_name="secondary", axis_label="Date")
        p_main_ax_x2.formatter = DatetimeTickFormatter(months=r"%Y/%m/%d",
                                                days=r"%Y/%m/%d %H:%M",
                                                hours=r"%Y/%m/%d %H:%M",
                                                minutes=r"%Y/%m/%d %H:%M")
        p_main.add_layout(p_main_ax_x2, 'above')

        p_main.add_tools(range_tool)

    # Create three subplots that show data in the selected range tool of the main plot

    pltDict =   {   
                    'ax1': {
                                'x':"x1", 
                                'y':"y1", 
                                "y_label": "mag" if not use_hostcorrected else "mag (corr)",
                                "err": ["y1_min", "y1_max"],                                  
                                # "marker": "circle", # better done
                                "marker": markermap,
                                "size": 5,
                                "color": index_cmap_selected,
                                "selected_color": index_cmap_selected,
                                "nonselected_color": index_cmap_nonselected,
                                "alpha": 0.5,
                                "selected_alpha": 0.5,
                                "nonselected_alpha": 0.5,
                            }, 
                    'ax2':  {
                                'x':"x1", 
                                'y':"y2", 
                                "y_label": "p [%]" if not use_hostcorrected else "p (corr) [%]",
                                "err": ["y2_min", "y2_max"],
                                # "marker":"circle",
                                "marker": markermap,
                                "size": 5,
                                "color": index_cmap_selected,
                                "selected_color": index_cmap_selected,
                                "nonselected_color": index_cmap_nonselected,
                                "alpha": 0.5,
                                "selected_alpha": 0.5,
                                "nonselected_alpha": 0.5,
                            },
                    'ax3':  {
                                'x':"x1", 
                                'y':"y3", 
                                "y_label": "chi [º]",
                                "err": ["y3_min", "y3_max"],                                  
                                # "marker":"circle",
                                "marker": markermap,
                                "size": 5,
                                "color": index_cmap_selected,
                                "selected_color": index_cmap_selected,
                                "nonselected_color": index_cmap_nonselected,
                                "alpha": 0.5,
                                "selected_alpha": 0.5,
                                "nonselected_alpha": 0.5,
                            },            
                }
                            
    for axLabel, axDict in pltDict.items():

        p = figure(title=None, x_range=selected_range, toolbar_location=None, tools=tools, lod_threshold=lod_threshold, lod_factor=lod_factor, lod_timeout=lod_timeout, output_backend="webgl", name=f"plot_{axLabel}")

        scatter_initial = Scatter(x=axDict['x'], y=axDict['y'], size=axDict["size"], fill_color=axDict["color"], line_color=axDict["color"], marker=axDict["marker"], fill_alpha=axDict["alpha"], line_alpha=axDict["alpha"])
        # scatter_selected = Scatter(x=axDict['x'], y=axDict['y'], size=axDict["size"], fill_color=axDict["selected_color"], line_color=axDict["selected_color"], marker=axDict["marker"], fill_alpha=axDict["selected_alpha"], line_alpha=axDict["selected_alpha"])
        scatter_nonselected = Scatter(x=axDict['x'], y=axDict['y'], size=axDict["size"], fill_color=axDict["nonselected_color"], line_color=axDict["nonselected_color"], marker=axDict["marker"], fill_alpha=axDict["nonselected_alpha"], line_alpha=axDict["nonselected_alpha"])
        pt_renderers = p.add_glyph(source, scatter_initial, selection_glyph=scatter_initial, nonselection_glyph=scatter_nonselected, view=view, name=f"{axLabel}_scatter_renderer")

        if enable_errorbars and axDict["err"] is not None:
            segs_initial = Segment(x0=axDict['x'], y0=axDict["err"][0], x1=axDict['x'], y1=axDict["err"][1], line_color=axDict["color"], line_alpha=axDict["alpha"])
            # segs_selected = Segment(x0=axDict['x'], y0=axDict["err"][0], x1=axDict['x'], y1=axDict["err"][1], line_color=axDict["selected_color"], line_alpha=axDict["selected_alpha"])
            segs_nonselected = Segment(x0=axDict['x'], y0=axDict["err"][0], x1=axDict['x'], y1=axDict["err"][1], line_color=axDict["nonselected_color"], line_alpha=axDict["nonselected_alpha"])
            errorbars_renderer = p.add_glyph(source, segs_initial, selection_glyph=segs_initial, nonselection_glyph=segs_nonselected, view=view, name=f"{axLabel}_errorbars_renderer")
            errorbars_renderer.visible = True

        if axLabel == "ax1":
            p.y_range.flipped = True
            p.extra_x_ranges["secondary"] = Range1d(*x2_range)
            p_ax_x2 = DatetimeAxis(x_range_name="secondary")
            p_ax_x2.formatter = DatetimeTickFormatter(months=r"%Y/%m/%d",
                                                days=r"%Y/%m/%d %H:%M",
                                                hours=r"%Y/%m/%d %H:%M",
                                                minutes=r"%Y/%m/%d %H:%M")
            p.add_layout(p_ax_x2, 'above')

        if axLabel == "ax3":
            p.below[0].axis_label = "MJD"

        if axLabel != "ax3":
            # clear x-axis of subplots:
            # p.below = [] # or pass x_axis_location="below" or x_axis_location=None to figure constructor
            # or leave the axis but remove markings:
            p.below[0].major_tick_line_color = None  # turn off x-axis major ticks
            p.below[0].minor_tick_line_color = None  # turn off x-axis minor ticks
            p.below[0].major_label_text_font_size = '0pt'  # turn off x-axis tick labels

        if axLabel == "ax2":
            p.yaxis.formatter = NumeralTickFormatter(format=' 0.0 %')
            #p.y_range.start = -0.01

        p.yaxis.axis_label = axDict["y_label"]
        p.title = axDict["y_label"]
        p.title.visible = False

        # Add common hover tool with custom JS callback
        hover_tool.renderers += [pt_renderers] # it has no renderers, append the scatter
        p.add_tools(hover_tool)

        # make box_zoom the default active tool
        p.add_tools(box_zoom_tool)
        p.toolbar.active_drag = box_zoom_tool

        axDict["p"] = p

    # Calculate the proportionality factor and shift between primary and secondary axes limits from the initial ranges
    # link the secondary and primary axes ranges in the p1 subplot
    # because the range_tool changes the p1 primary axes but not the secondary (otherwise they are already linked)

    p1, p2, p3 = pltDict["ax1"]["p"], pltDict["ax2"]["p"], pltDict["ax3"]["p"]
    
    factor = (x2_lims[1] - x2_lims[0]) / (x1_lims[1] - x1_lims[0])
    shift = x2_lims[0] - x1_lims[0]*factor

    callback = CustomJS(args=dict(primary=p1.x_range, secondary=p1.extra_x_ranges["secondary"], factor=factor, shift=shift), code = """secondary.start = factor * primary.start + shift; secondary.end = factor * primary.end + shift;""")
    p1.x_range.js_on_change('start', callback)
    p1.x_range.js_on_change('end', callback)

    # Combine plots in a column layout

    if enable_full_lc:
        plot_layout = gridplot([[p_main], [pltDict["ax1"]["p"]], [pltDict["ax2"]["p"]], [pltDict["ax3"]["p"]]], merge_tools=True, toolbar_location="right", sizing_mode='stretch_both')
    else:
        plot_layout = gridplot([[pltDict["ax1"]["p"]], [pltDict["ax2"]["p"]], [pltDict["ax3"]["p"]]], merge_tools=True, toolbar_location="right", sizing_mode='stretch_both')

    plot_layout.name = "gridplot"

    # this is a fix for the save tool of the gridplot not behaving as expected
    # and having the filename of the configured plot
    for tool in plot_layout.toolbar.tools:
        if isinstance(tool, SaveTool):
            tool.filename = save_filename
            break

    # Add a callback to hide errorbars when panning (it makes the plot smoother)

    if enable_errorbars:
        cb_hide_errorbars = """window.were_errorbars_active = document.querySelector('#cbox_errorbars').checked; plot_hide_errorbars();"""
        cb_restore_errorbars = """if (window.were_errorbars_active) {plot_show_errorbars();}"""

        if enable_full_lc:
            p_main.js_on_event(PanStart, CustomJS(code=cb_hide_errorbars))
            p_main.js_on_event(PanEnd, CustomJS(code=cb_restore_errorbars))

        p1.js_on_event(PanStart, CustomJS(code=cb_hide_errorbars))
        p1.js_on_event(PanEnd, CustomJS(code=cb_restore_errorbars))
        p2.js_on_event(PanStart, CustomJS(code=cb_hide_errorbars))
        p2.js_on_event(PanEnd, CustomJS(code=cb_restore_errorbars))
        p3.js_on_event(PanStart, CustomJS(code=cb_hide_errorbars))
        p3.js_on_event(PanEnd, CustomJS(code=cb_restore_errorbars))
        
    # add a callback to anable flagging data
    
    cb_stage_to_flag = CustomJS(args=dict(source=source), code="""
        vueApp.$data.selected_plot_idx = source.selected.indices;
    """)
    source.selected.js_on_change('indices', cb_stage_to_flag)

    ##############
    # Data table #
    ##############
    
    table_link_formatter = HTMLTemplateFormatter(template="""<a href="/iop4/admin/iop4api/photopolresult/?id=<%= value %>"><%= value %></a>""")
    float_formatter = NumberFormatter(format='0.00')
    percent_formatter = NumberFormatter(format='0.0 %')
    flags_to_str_formatter = HTMLTemplateFormatter(template="""<span><%= flags_to_str(flags) %></span>""")
    
    table_columns = [   
                        TableColumn(field="pk", title="id", formatter=table_link_formatter),
                        TableColumn(field="instrument", title="instrument"),
                        TableColumn(field="x1", title="mjd", formatter=float_formatter),
                        TableColumn(field="y1", title="mag", formatter=float_formatter),
                        TableColumn(field="y1_err", title="dmag", formatter=float_formatter),
                        TableColumn(field="y2", title="p", formatter=percent_formatter),
                        TableColumn(field="y2_err", title="dp", formatter=percent_formatter),
                        TableColumn(field="y3", title="chi", formatter=float_formatter),
                        TableColumn(field="y3_err", title="dchi", formatter=float_formatter),
                        TableColumn(field="flags", title="flags", formatter=flags_to_str_formatter),
                    ]
               
    data_table = DataTable(source=source, view=view, columns=table_columns, index_position=None,
                           sortable=True, reorderable=True, scroll_to_selection=True, #sizing_mode="stretch_width", # this works and is fast
                            sizing_mode="inherit", # this also works fast while respecting the container size
                            stylesheets=[InlineStyleSheet(css=""".slick-cell.selected { background-color: #d2eaff; }""")]) 

    #################################################
    # Create a legend (it will be a different plot) #
    #################################################

    # Create a ColumnDataSource with data for Circle and Star glyphs

    legend_source = ColumnDataSource(data=dict(x=[0], y=[0]))

    label_stylesheet = InlineStyleSheet(css="""
                                        :host { 
                                            display: flex;
                                            align-items: self-end;
                                            margin: 0;
                                            margin-left: 10px;
                                            min-width: 6em;
                                            /* max-width: 12em; */
                                        }

                                        label {
                                            display: flex;
                                            cursor: pointer; 
                                            font-weight: normal;
                                            white-space: nowrap;
                                        }

                                        span {
                                            display: flex;
                                            align-items: flex-end;
                                            min-width: 4em;
                                        }

                                        input {
                                            cursor: pointer;
                                        }

                                        label:active {
                                            background-color: #f2f2f2;
                                            border-radius: 4px;
                                        }
                                        """)

    legend_row_L = list()

    # Instrument legend entries

    for instrument, color in zip(instruments_L, instrument_color_L):
        # Create Div elements for labels
        label = Div(text=f"""<label><span>{instrument}</span><input onclick="plot_hide_instrument(this);" data-instrument="{instrument}" type="checkbox" checked/></label>""", height=21, stylesheets=[label_stylesheet])
        # Create glyphs
        circle_glyph = Circle(x='x', y='y', fill_color=color, line_color=color, size=8)
        # Create plots to hold the glyphs
        legend_p = figure(width=21, height=21, toolbar_location=None, min_border=0, tools="")
        legend_p.add_glyph(legend_source, circle_glyph)
        legend_p.axis.visible = False
        legend_p.grid.visible = False
        legend_p.background_fill_alpha = 0.0
        legend_p.outline_line_alpha = 0.0

        # Combine glyphs and labels into rows
        legend_row = Row(children=[legend_p, label], align="center")

        legend_row_L.append(legend_row)

    # Flag legend entries
    # values here must match those in PhotoPolResult.FLAGS (0, 1 << 1 = 2, 1 << 2 = 4, 1 << 1 | 1 << 2 = 6)
        
    for flag_value, flag_label, marker in zip([2,4,6], ["bad photometry", "bad polarimetry", "bad photometry and polarimetry"], ["triangle", "inverted_triangle", "cross"]):
        # Create Div elements for labels
        if flag_value in [2,4]:
            cbox = f"""<input onclick="plot_hide_flag(this);" data-flag="{flag_value}" type="checkbox" checked/>"""
            no_pointer = ""
        else:
            cbox = ""
            no_pointer = """ style="cursor: default;" """
        label = Div(text=f"""<label {no_pointer}><span>{flag_label}</span>{cbox}</label>""", height=21, stylesheets=[label_stylesheet])
        # Create glyphs
        circle_glyph = Scatter(x='x', y='y', fill_color="black", line_color="black", size=8, marker=marker)
        # Create plots to hold the glyphs
        legend_p = figure(width=21, height=21, toolbar_location=None, min_border=0, tools="")
        legend_p.add_glyph(legend_source, circle_glyph)
        legend_p.axis.visible = False
        legend_p.grid.visible = False
        legend_p.background_fill_alpha = 0.0
        legend_p.outline_line_alpha = 0.0

        # Combine glyphs and labels into rows
        legend_row = Row(children=[legend_p, label], align="center")

        legend_row_L.append(legend_row)

    # Combine rows into a column and display
    legend_layout = Column(children=legend_row_L)




    # Get the components to embed in the Django template

    doc = Document()
    doc.add_root(plot_layout)
    doc.add_root(data_table)
    doc.add_root(legend_layout)
    
    from bokeh.embed.util import standalone_docs_json_and_render_items
    doc, render_items = standalone_docs_json_and_render_items([plot_layout, data_table, legend_layout])
    return JsonResponse({'doc':doc,
                        'render_items':[item.to_json() for item in render_items],
                        'n_points':len(x1),
                        'enable_full_lc': enable_full_lc,
                        'enable_crosscheckpts': enable_crosscheckpts,
                        'enable_errorbars': enable_errorbars,
                        'use_hostcorrected': use_hostcorrected})